#' PerformMainFit
#'
#' Perform main fit
#'
#' @param context List of parameters. Required.
#' @param data Input data as data.table. Required.
#' @param param List of parameters. Optional. Default = NULL.
#' @param info List of parameters. Optional. Default = NULL.
#' @param maxNoFit Maximum number of amoeba iterations. Optional. Default = 30.
#' @param ctol Minium required deviance in consecutive lambda estimations.
#'   Optional. Default = 1e-6.
#' @param ftol Minium required deviance in amoeba calculations. Optional.
#'   Default = 1e-5.
#' @param algorithm Name of optimization algorithm from package \code{nloptr} to use for bootstrap
#'   iterations. Default = 'NLOPT_LN_BOBYQA'
#' @param verbose Logical indicating to print detailed info during fitting. Optional.
#'   Default = \code{FALSE}
#'
#' @return
#' Results list object
#'
#' @examples
#' \dontrun{
#' PerformMainFit(
#'   context, data, param, info, maxNoFit = 2, ctol = 1e-6, ftol = 1e-5, verbose = TRUE
#' )
#' }
#'
#' @export
PerformMainFit <- function(
  context,
  data,
  param = NULL,
  info = NULL,
  maxNoFit = 30L,
  ctol = 1e-6,
  ftol = 1e-5,
  algorithm = 'NLOPT_LN_BOBYQA',
  verbose = FALSE
) {
  if (is.null(info) || is.null(param)) {
    runType <- 'MAIN'
    info <- GetInfoList(context)
    param <- GetParamList(context, info)
    message('Run type: ', runType, ' - all initial parameters will be determined from context')
  } else {
    runType <- 'MAIN_WITH_INIT'
    message(
      'Run type: ',
      runType,
      ' - all initial parameters will be determined from provided "info" and "param" objects'
    )
  }
  probSurv1996 <- GetProvSurv96(param, info)

  tmpModelFitDist <- info$ModelFitDist
  info$ModelFitDist <- 'POISSON'
  if (tmpModelFitDist != info$ModelFitDist) {
    message('Input distribution was set to "', tmpModelFitDist, '. ',
            'This is overridden to "', info$ModelFitDist, '" for the main fit.')
  }

  converged <- FALSE
  while (!converged) {
    nTheta <- 100
    while (nTheta != param$NoTheta) {
      message(sprintf('Number of estimated spline weights: %d', param$NoTheta))
      nTheta <- param$NoTheta

      res <- EstimateParameters(
        runType = runType,
        probSurv1996, param, info, data,
        maxNoFit, ctol, ftol, verbose
      )

      p <- res$P
      converged <- res$Converged
      param <- res$Param
      info <- res$Info

      selSmallTheta <- abs(param$Theta) < 1
      param$ThetaP[selSmallTheta] <- 0
      param$Theta[selSmallTheta] <- 0
      param$NoTheta <- sum(param$ThetaP)

      iterResults <- res$IterResults
      lastResults <- res$IterResults[[length(res$IterResults)]]
    }

    if (!converged) {
      message(sprintf('Fit did NOT converge, goodness-of-fit: %f\n', lastResults$LLTotal))
      context$Parameters$Models$INCIDENCE$ModelNoKnots <- info$ModelNoKnots - 1
      info <- GetInfoList(context)
      param <- GetParamList(context, info)
      probSurv1996 <- GetProvSurv96(param, info)
    } else {
      message(sprintf('Fit converged, goodness-of-fit: %f\n', lastResults$LLTotal))
    }
  }

  message(sprintf('beta[%d]: %f\n', seq_len(param$NoDelta), param$Beta[seq_len(param$NoDelta)]))
  message(sprintf(
    'theta[%d]: %f\t- %s\n',
    seq_along(param$Theta),
    param$Theta,
    ifelse(param$ThetaP, 'FIXED', 'NOT FIXED')
  ))

  info$ModelFitDist <- tmpModelFitDist
  # Estimate overdispersion for negative binomial
  if (info$ModelFitDist == 'NEGATIVE_BINOMIAL') {
    message('Estimating overdispersion type ', info$OverdisperionType)

    rMin <- 1.0
    rMax <- 100000

    if (info$OverdisperionType != 2) {
      stop(sprintf('Overdisperion type %d is not supported', info$OverdisperionType))
    } else if (info$OverdisperionType == 2) {
      extraArgs <- list(
        Type = 0,
        ModelResults = lastResults$ModelResults,
        Data = data,
        Info = info,
        Param = param
      )
      # Use base R function 'uniroot' instead of zbrent
      param$RDispAIDS <- zbrent(FitLLrAIDS, rMin, rMax, ftol, extraArgs)
      param$RDispRest <- zbrent(FitLLrRest, rMin, rMax, ftol, extraArgs)

      message(sprintf('Overdispersion: AIDS = %f, Rest = %f', param$RDispAIDS, param$RDispRest))
    }
  }

  res <- FitLLTotal(p, probSurv1996, param, info, data)
  modelResults <- res$ModelResults
  statRes <- FitStatistics(modelResults, info, data, param)

  countResults <- ModelCountResults(modelResults, info, param)
  timeResults <- ModelTimeResults(modelResults, info, param)
  mainOutputs <- ModelOutputs(modelResults, countResults, timeResults, info, param, data)

  return(list(
    Converged = converged,
    P = p,
    Info = info,
    Param = param,
    Statistics = statRes,
    IterResults = iterResults,
    ModelResults = modelResults,
    CountResults = countResults,
    TimeResults = timeResults,
    MainOutputs = mainOutputs
  ))
}
